/*
 * Copyright (c) 2005
 *    Christian Biere <christianbiere@gmx.de> All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "server.h"

#ifdef USE_POLL

#ifdef USE_KQUEUE
#error USE_KQUEUE and USE_POLL defined!
#endif

#else

#if defined(HAVE_KQUEUE) && !defined(USE_KQUEUE)
#define USE_KQUEUE
#endif /* USE_KQUEUE */

#ifndef USE_KQUEUE
#define USE_POLL
#endif /* !USE_KQUEUE*/

#endif /* USE_POLL */

#ifdef USE_POLL
#include <poll.h>

/* If still not defined, just define them */
#ifndef POLLRDNORM
#define POLLRDNORM 0
#endif
#ifndef POLLRDBAND
#define POLLRDBAND 0
#endif
#ifndef POLLWRNORM
#define POLLWRNORM 0
#endif
#ifndef POLLWRBAND
#define POLLWRBAND 0
#endif
#endif /* USE_POLL */

#include "connection.h"
#include "oop.h"

#define SERVER_CHUNKS 64

/* XXX: "server" is really a poor name for this */

typedef struct server_private {
  struct server         vtable;

  uint64_t              magic;
  connection_t          **clients;
  size_t                size_clients;
  size_t                num_clients;
  size_t                num_closed;
  server_periodic_cb_t  periodic_cb;
  unsigned long         timeout;  /* delay for periodic callbacks */
  time_t                last_gc;
  
#ifdef USE_KQUEUE
  struct kevent         *kevents;
  int kq;
#endif /* USE_KQUEUE */

#ifdef USE_POLL
  struct pollfd         *pfds;
#endif /* USE_POLL */

} server_private_t;

typedef union {
  server_t pub;
  server_private_t priv;
} server_union;

static const uint64_t server_magic =
  MAGIC_CHARS('S', 'E', 'R', 'V', 'E', 'R', '_', 'T');

#define PRIV(x) (((server_union *) (x))->priv)

ATTRIBUTE_SET(server, unsigned long, timeout)

static void
server_set_periodic_cb(server_t *server, server_periodic_cb_t cb)
{
  server_private_t *s = &PRIV(server);

#ifdef USE_KQUEUE
  static const struct timespec ts;
  struct kevent kev;

  EV_SET(&kev, (uintptr_t) server,
      EVFILT_TIMER, (cb ? EV_ADD : EV_DELETE), 0, s->timeout, 0);
  if (-1 == kevent(s->kq, &kev, 1, 0, 0, &ts)) {
    WARN("kevent() failed: %s", compat_strerror(errno));
  }
#endif /* USE_KQUEUE */
  
  s->periodic_cb = cb;
}
  
static void
server_destruct(server_t *server)
{
  server_private_t *s = &PRIV(server);

#ifdef USE_KQUEUE
  if (s->kq != -1) {
    close(s->kq);
  }
  DO_FREE(s->kevents);
#endif /* USE_KQUEUE */
  
#ifdef USE_POLL
  DO_FREE(s->pfds);
#endif /* USE_POLL */
  
  memset(s, 0, sizeof *s);
  DO_FREE(s);
}

static inline bool
server_should_collect_garbage(server_private_t *s, time_t now)
{
  return 0 != s->num_closed || 0 != difftime(now, s->last_gc);
}

static void
server_garbage_collect(server_private_t *s, time_t now)
{
  bool timeouts;
  size_t i, j;

  timeouts = 0 != difftime(now, s->last_gc);
  if (!timeouts && 0 == s->num_closed)
    return;

  s->last_gc = now;

  for (i = 0; i < s->size_clients; i++) {
    connection_t *c;
    unsigned int it;
    int fd;

    assert(i < s->size_clients);
    if (NULL == (c = s->clients[i]))
      continue;

    fd = c->get_fd(c);
    if (timeouts && fd >= 0) {
      it = c->get_maxidle(c);
      if (it && ((unsigned long) difftime(now, c->get_stamp(c)) >= it)) {
        c->event(c, CONN_EV_TIMEOUT, now);
        fd = c->get_fd(c);
        if (-1 == fd)
          s->num_closed++;
      }
    }

    if (fd < 0) {
      assert(fd == -1);
      assert(s->num_closed > 0);
      DESTRUCT(c);
      s->clients[i] = NULL;
      s->num_clients--;
      s->num_closed--;
    }

  }

  j = 0;
  for (i = 0; i < s->size_clients; i++) {
    if (!s->clients[i]) {
      while (j < s->size_clients && !s->clients[j])
        j++;

      if (j >= s->size_clients)
        break;

      s->clients[i] = s->clients[j];
      s->clients[j] = NULL;
    }
  }

}

static void
server_watch_client(server_t *server, connection_t *c, connection_event_t ev)
{
  server_private_t *s = &PRIV(server);
  
#ifdef USE_KQUEUE
  struct kevent kev[2];
  size_t n = 0;
#endif /* USE_KQUEUE */
  
  connection_event_t oev;
  int fd;
  bool r = false, w = false;

  assert(c != NULL);
  assert(s->num_clients > 0);
  fd = c->get_fd(c);
  assert(fd >= 0);

  oev = c->get_eventmask(c);
  if (oev == ev)
    return;
  
  switch (ev) {
  case CONN_EV_READ:
    r = true;
    break;
  case CONN_EV_WRITE:
    w = true;
    break;
  case CONN_EV_NONE:
    break;
  default:
    assert(0);
  }

#ifdef USE_KQUEUE
  if (oev != CONN_EV_NONE) {
    assert(oev == CONN_EV_WRITE || oev == CONN_EV_READ);
    
    EV_SET(&kev[n], fd, (oev == CONN_EV_READ ? EVFILT_READ : EVFILT_WRITE),
      EV_DELETE | EV_DISABLE, 0, 0, (uintptr_t) c);
    n++;
  }

  if (r || w) {
    EV_SET(&kev[n], fd, (r ? EVFILT_READ : EVFILT_WRITE),
      EV_ADD, 0, 0, (uintptr_t) c);
    n++;
  }

  if (n != 0) {
    static const struct timespec ts;

    assert(n <= ARRAY_LEN(kev));
    if (-1 == kevent(s->kq, kev, n, NULL, 0, &ts)) {
      WARN("kevent() failed: %s", compat_strerror(errno));
    }
  }
#endif /* USE_KQUEUE */

  c->set_eventmask(c, ev);
}

/**
 * This function is only used with kqueue() and because up to NetBSD 2.0H
 * closing a pipe doesn't cause a EVFILT_READ in a remote process reading
 * from the the pipe. Therefore, the other process would wait forever
 * instead of terminating automagically after the remote process.
 */
static void
server_watch_process(server_t *server, pid_t pid, server_process_cb_t cb)
{
  server_private_t *s = &PRIV(server);
  
#ifdef USE_KQUEUE
  static const struct timespec ts;
  struct kevent kev;

  EV_SET(&kev, pid, EVFILT_PROC, cb != NULL ? EV_ADD : EV_DELETE,
      0, NOTE_EXIT, (uintptr_t) cb);
  if (-1 == kevent(s->kq, &kev, 1, NULL, 0, &ts)) {
    WARN("kevent() failed: %s", compat_strerror(errno));
  }
#endif /* USE_KQUEUE */

#ifdef USE_POLL
  (void) s;
  (void) pid;
  (void) cb;
#endif /* USE_POLL */
}


#ifdef USE_POLL
typedef enum {
  POLL_READ = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI,
  POLL_WRITE = POLLOUT | POLLWRNORM | POLLWRBAND,
  POLL_ANY = POLL_READ | POLL_WRITE
} poll_event_t;

static inline int
conn_ev_to_poll(connection_event_t ev)
{
  return  (ev & CONN_EV_READ ? POLL_READ : 0) |
          (ev & CONN_EV_WRITE ? POLL_WRITE : 0);
}

static void
server_create_pollset(server_private_t *s)
{
  size_t i, j;
 
  i = 0;
  for (j = 0; j < s->size_clients && i < s->num_clients; j++) {
    connection_t *c;
    struct pollfd *pfd;
   
    c = s->clients[j];
    if (!c)
      continue;

    assert(c != NULL);
    assert(i < s->num_clients);
    pfd = &s->pfds[i++];

    pfd->fd = c->get_fd(c);
    assert(pfd->fd >= 0);

    pfd->events = conn_ev_to_poll(c->get_eventmask(c));
  }
}

static void
server_inform_clients(server_private_t *s, time_t now)
{
  size_t i, j;

  i = 0;
  for (j = 0; j < s->size_clients && i < s->num_clients; j++) {
    connection_t *c;
    struct pollfd *pfd;
    connection_event_t ev;

    c = s->clients[j];
    if (!c)
      continue;

    assert(i < s->num_clients);
    pfd = &s->pfds[i++];
  
    if (c->get_fd(c) < 0) {
      /* This happens when an incoming connection was accept()ed and
       * the incoming cb handler was called which closed the
       * connection. */
      assert(-1 == c->get_fd(c));
      CRIT("Invalid file descriptor");
      s->num_closed++;
      continue;
    }
    
    if (c->get_fd(c) != pfd->fd) {
      CRIT("file descriptor mismatch; c->fd=%d; s->pfds[i].fd=%d",
        c->get_fd(c), pfd->fd);
      assert(c->get_fd(c) == pfd->fd);
    }
    
    if (pfd->revents != 0) {
      ev = CONN_EV_NONE;
      if (pfd->revents & POLL_READ)
        ev |= CONN_EV_READ;
      if (pfd->revents & POLL_WRITE)
        ev |= CONN_EV_WRITE;
      if (pfd->revents & POLLHUP)
        ev |= CONN_EV_HANGUP;
      if (pfd->revents & POLLERR) {
        ev |= CONN_EV_ERROR;
      }
      
      if (pfd->revents & POLLNVAL) {
        ev |= CONN_EV_ERROR;
        WARN("POLLNVAL (fd=%d)", pfd->fd);
      }

      if (ev != CONN_EV_NONE) {
        c->event(c, ev, now);
        if (-1 == c->get_fd(c))
          s->num_closed++;
      }
    }
  }

  server_garbage_collect(s, now);
}

static inline int
server_periodic_handler(server_private_t *s, const struct timeval *tv)
{
  static struct timeval last_tv;
  
  if (!s->periodic_cb)
    return 0;

  if (!last_tv.tv_sec)
    last_tv = *tv;
    
  if (DIFFTIMEVAL(tv, &last_tv) >= s->timeout * 1000) {
    s->periodic_cb((server_t *) s, tv->tv_sec);
    last_tv = *tv;
    return 1;
  }

  return 0;
}
#endif /* USE_POLL */

static int
server_add_client(server_t *server, connection_t *c)
{
  server_private_t *s = &PRIV(server);
  void *p;
  size_t size, i;
  
#ifdef USE_KQUEUE
  p = s->kevents;
  size = sizeof s->kevents[0];
#endif /* USE_KQUEUE */
#ifdef USE_POLL
  p = s->pfds;
  size = sizeof s->pfds[0];
#endif /* USE_POLL */

  assert(c != NULL);
  assert(c->get_fd(c) >= 0);

  if (s->num_clients == s->size_clients) {
    size_t n;

    n = s->size_clients;
    n = n < SERVER_CHUNKS ? SERVER_CHUNKS : n * 2;
    p = realloc(p, n * size);
    if (!p)
      return -1;
  }
  assert(p != NULL);

#ifdef USE_KQUEUE
  s->kevents = p;
  memset(&s->kevents[s->num_clients], 0, sizeof s->kevents[0]);
#endif /* USE_KQUEUE */

#ifdef USE_POLL
  s->pfds = (struct pollfd *) p;
  s->pfds[s->num_clients].fd = c->get_fd(c);
  s->pfds[s->num_clients].events = 0;
  s->pfds[s->num_clients].revents = 0;
#endif /* USE_POLL */

  /*
   * This is OK when using poll() because the new client is appended
   * so it's not a problem when a event callback adds a client while
   * being in server_inform_clients().
   */
 
  if (s->num_clients == s->size_clients) {
    size_t n;
    
    n = s->size_clients;
    n = n < SERVER_CHUNKS ? SERVER_CHUNKS : n * 2;
    p = realloc(s->clients, n * sizeof s->clients[0]);
    if (!p)
      return -1;

    s->clients = p;
    for (i = s->size_clients; i < n; i++) {
      s->clients[i] = NULL;
    }
    s->size_clients = n;
  }
  
  assert(s->num_clients < s->size_clients);
  for (i = 0; i < s->size_clients; i++) {
    if (!s->clients[i]) {
      s->clients[i] = c;
      break;
    }
  }
  assert(i < s->size_clients);
  s->num_clients++;

  return 0;
}

#ifdef USE_KQUEUE
static void
server_mainloop(server_t *server)
{
  server_private_t *s = &PRIV(server);

  assert(s->kq != -1);
  set_signal(SIGPIPE, SIG_IGN);
  for (;;) {
    int ret;

    ret = kevent(s->kq, NULL, 0, s->kevents, s->num_clients, NULL);
    if (ret == -1) {
      WARN("kevent() failed: %s", compat_strerror(errno));
    } else if (ret > 0) {
      size_t i;
      time_t now;
      bool periodic = false;
      
      assert(ret >= 0 && (size_t) ret <= s->num_clients);
      now = compat_mono_time(NULL);
      
      for (i = 0; i < (size_t) ret; i++) {
#if 0
         DBUG("kevent(): ident=%p filter=%lu flags=%lu fflags=%lu "
           "data=%" PRId64 " udata=%p %s",
           (void *) s->kevents[i].ident,
           (unsigned long) s->kevents[i].filter,
           (unsigned long) s->kevents[i].flags,
           (unsigned long) s->kevents[i].fflags,
           s->kevents[i].data,
           (void *) s->kevents[i].udata,
           s->kevents[i].flags & EV_EOF ? "EOF" : ""
         );
#endif

        if (s->kevents[i].filter == EVFILT_TIMER) {
          assert(s->kevents[i].ident == (uintptr_t) s);
          periodic = true;
        } else if (s->kevents[i].filter == EVFILT_PROC) {
          server_process_cb_t cb;
          pid_t pid;

          pid = s->kevents[i].ident;
          cb = (server_process_cb_t) s->kevents[i].udata;
          assert(cb != NULL);
          cb(server, pid);
        } else {
          int fd, n;
          connection_t *c;
          connection_event_t ev;

          fd = s->kevents[i].ident;
          assert(fd >= 0);
          c = (connection_t *) s->kevents[i].udata;
          assert(c != NULL);
          assert(c->get_fd(c) == fd);
       
          ev = CONN_EV_NONE;
          n = 1;
          switch (s->kevents[i].filter) {
          case EVFILT_READ:
            ev = CONN_EV_READ;
            if (c->get_listening(c)) {
              n = MIN(s->kevents[i].data, 16);
            }
            if (s->kevents[i].data == 0)
              ev |= CONN_EV_HANGUP;
            break;
          case EVFILT_WRITE:
            ev = CONN_EV_WRITE;
            if (s->kevents[i].data == 0)
              ev |= CONN_EV_HANGUP;
            break;
          default:
            assert(0);
          }
        
          while (n-- > 0 && (ev & c->get_eventmask(c))) {
            c->event(c, ev, now);
            fd = c->get_fd(c);
            if (-1 == fd) {
              s->num_closed++;
              break;
            }
          }

        }
 
      }

      server_garbage_collect(s, now);
      
      if (periodic && s->periodic_cb) {
        s->periodic_cb(server, now);
      }
    
    } else if (0 == s->num_clients && 0 != s->timeout) {
      struct kevent kev;

      ret = kevent(s->kq, NULL, 0, &kev, 1, NULL);
      switch (ret) {
      case 0: break;
      case -1:
        WARN("kevent() failed: %s", compat_strerror(errno));
        break;
      default:  
        if (kev.filter == EVFILT_TIMER) {
          assert(kev.ident == (uintptr_t) s);
          if (s->periodic_cb)
            s->periodic_cb(server, compat_mono_time(NULL));
        }
      }
    }
  }
  
}

#endif /* USE_KQUEUE */

#ifdef USE_POLL
static void
server_mainloop(server_t *server)
{
  server_private_t *s = &PRIV(server);

  set_signal(SIGPIPE, SIG_IGN);
  for (;;) {
    struct timeval tv;
    int ret;

    server_create_pollset(s);

    do {
      ret = poll(s->pfds, s->num_clients, s->timeout);
      if (-1 == ret && errno != EINTR && errno != EAGAIN) {
        WARN("poll failed: %s", compat_strerror(errno));
        sleep(1); /* prevent exhausting CPU resources */
      }

      compat_mono_time(&tv);
      if (server_periodic_handler(s, &tv))
        break;
    } while (ret == 0);
    
    server_inform_clients(s, tv.tv_sec);
  }
}

#endif /* USE_POLL */

static int
server_error(server_t *s, const char *where, const char *what, int error)
{
  (void) s;
  fprintf(stderr, "%s: %s%s\n", where, what, compat_strerror(errno));
  return error;
}

static const struct server server_vtable = {
  &server_mainloop,
  &server_error,
  &server_add_client,
  &server_watch_client,
  &server_watch_process,
  &server_set_periodic_cb,
  &server_set_timeout,
  &server_destruct
};


server_t *
server_new(void)
{
  server_private_t *s = (server_private_t *) calloc(1, sizeof *s);

  if (s) {
    s->vtable = server_vtable;
    s->magic = server_magic;
    s->clients = NULL;
    s->size_clients = 0;
    s->num_clients = 0;
    s->periodic_cb = NULL;
    s->last_gc = compat_mono_time(NULL);
    s->num_closed = 0;

#ifdef USE_POLL
    s->pfds = NULL;
#endif /* USE_POLL */
    
#ifdef USE_KQUEUE
    s->kevents = NULL;
    s->kq = kqueue();
    if (-1 == s->kq) {
      free(s);
      return NULL;
    }
#endif /* USE_KQUEUE */
    
  }
  return (server_t *) s;
}


/* vi: set ai et sts=2 sw=2 cindent: */
