/*
 * Copyright (c) 2004
 *    Christian Biere <christianbiere@gmx.de> All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "udp.h"

#include "lib/append.h"
#include "lib/ggep.h"
#include "lib/guid.h"
#include "lib/list.h"
#include "lib/nettools.h"
#include "lib/mem.h"

static connection_t *udp_con;
static void (*udp_pong_cb)(const net_addr_t *addr, in_port_t, void *);

typedef enum {
  GNET_P_PING = 0x00,
  GNET_P_PONG = 0x01,
  GNET_P_PUSH = 0x40,
  GNET_P_QASK = 0x80,
  GNET_P_QHIT = 0x81
} gnet_p_t;

typedef struct ping_info {
  guid_t guid;
  struct timeval tv;
  net_addr_t addr;
  in_port_t port;
  bool replied;
} ping_info_t;

static mem_cache_t *mc_ping_info;

static void
udp_pong_callback(const net_addr_t addr, in_port_t port, void *udata)
{
  if (udp_pong_cb)
    udp_pong_cb(&addr, port, udata);
}

void
udp_set_pong_callback(void (*cb)(const net_addr_t *, in_port_t, void *))
{
  udp_pong_cb = cb;
}

static void
udp_send_packet(const net_addr_t addr, in_port_t port,
    const void *data, size_t size)
{
  const struct sockaddr *to;
  socklen_t len;
  ssize_t ret;

  assert(udp_con);
  assert(data);

  len = net_addr_sockaddr(addr, port, &to);
  ret = sendto(udp_con->get_fd(udp_con), data, size, 0, to, len);
  if (ret == (ssize_t) -1) {
    DBUG("sendto() failed: %s", compat_strerror(errno));
  } else if (ret == 0) {
    DBUG("sendto() returned zero");
  } else if ((size_t) ret != size) {
    DBUG("sendto() partial write: ret=%d", (int) ret);
  } else {
#if 0
    DBUG("sendto() succeeded");
#endif
  }
}

void
udp_send_ping(net_addr_t addr, in_port_t port, bool with_ggep_scp)
{
  static const char ping_data[] = {
    /* 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,*/ /* GUID */
    GNET_P_PING,  /* Function (PING) */
    1,            /* TTL */
    0,            /* Hops */
    0, 0, 0, 0,   /* Size (corrected before sent) */
  };
  struct timeval tv;
  char buf[4096], *p = buf;
  size_t left = sizeof buf, size, len;
  ping_info_t *pi;
  void *udata;

  STATIC_ASSERT(23 - 16 == sizeof ping_data);
  STATIC_ASSERT(16 == sizeof pi->guid);

  assert(udp_con != NULL);

  if (NULL == (pi = mem_alloc(mc_ping_info))) {
    CRIT("mem_alloc() failed");
    return;
  }

  gettimeofday(&tv, NULL);
  guid_create(&pi->guid);
  pi->addr = addr;
  pi->port = port;
  pi->tv = tv;
  pi->replied = false;

#if 0
  {
    char addr_buf[NET_ADDR_PORT_BUFLEN];
    
    print_net_addr_port(addr_buf, sizeof addr_buf, pi->addr, pi->port);
    DBUG("Target: %s\nPing GUID: %02x %02x %02x %02x %02x %02x %02x %02x - "
        "%02x %02x %02x %02x %02x %02x %02x %02x",
        addr_buf,
        pi->guid.u8[0x0], pi->guid.u8[0x1], pi->guid.u8[0x2], pi->guid.u8[0x3],
        pi->guid.u8[0x4], pi->guid.u8[0x5], pi->guid.u8[0x6], pi->guid.u8[0x7],
        pi->guid.u8[0x8], pi->guid.u8[0x9], pi->guid.u8[0xa], pi->guid.u8[0xb],
        pi->guid.u8[0xc], pi->guid.u8[0xd], pi->guid.u8[0xe], pi->guid.u8[0xf]);
  }
#endif

  if (guid_is_magic(&pi->guid, &udata)) {
    raise(SIGTRAP);
    mem_free(mc_ping_info, pi);
    CRIT("Oops GUID collision");
    return;
  }

  if (guid_add(&pi->guid, pi)) {
    mem_free(mc_ping_info, pi);
    CRIT("guid_add() failed");
    return;
  }

  /* 128-bit Message ID */
  p = append_chars(p, &left, &pi->guid.u8[0], 16);
  /* Pseudo ping header */
  p = append_chars(p, &left, ping_data, sizeof ping_data);

  /* Payload follows */
  if (with_ggep_scp) {
    uint8_t pref;
    ggep_t gtx;

    (void) ggep_init(&gtx, p, left);
    
    /* LimeWire 4.2.2 doesn't reply to SCP pings without a preference
     * for either ultrapeers or leaves
     */ 
    
    pref = 1; /* Prefer for ultrapeers */
    ggep_pack(&gtx, GGEP_ID_SCP, 0, &pref, 1);
    len = ggep_end(&gtx);
    assert(left >= len);
    left -= len;
  }
   
  assert(left <= sizeof buf);
  size = sizeof buf - left;
  assert(size >= 23);
  
  len = size - 23;  /* Subtract the header size */
  assert(len < sizeof buf);

  /* Correct the payload length field of the Gnutella header */
  poke_le16(&buf[19], len);

  udp_send_packet(addr, port, buf, size);
}

static void
handle_packet(connection_t *c, uint8_t *data, size_t len,
    const net_addr_t sender_addr, in_port_t sender_port)
{
  unsigned int function, ttl, hops, size;
  uint8_t *p = data;
  size_t data_len;
  char id_name[GGEP_ID_BUFLEN];
  char addr_buf[NET_ADDR_PORT_BUFLEN];
  ggep_t gtx;
  ggep_id_t id;
  net_addr_t pong_addr = net_addr_unspecified;
  in_port_t pong_port = 0;
  char vendor[5] = { 0, 0, 0, 0, 0 };
  bool host_is_udphc = false;
  bool has_scp = false;
  bool has_vc = false;  /* set if there was a GGEP VC block */
  uint32_t vc = 0;  /* vendor code */
  struct peer_set ipp_peers;

  (void) c;

  ipp_peers.n = 0;
    
  if (len < 23) {
    DBUG("Packet is too small");
    return;
  }

  function = data[16];
  ttl = data[17];
  hops = data[18];
  size = peek_le32(&data[19]);

#if 0
  DBUG("GUID: %02x %02x %02x %02x %02x %02x %02x %02x - "
    "%02x %02x %02x %02x %02x %02x %02x %02x\n"
    "function: %02x, ttl: %3d, hops: %3d, size: %d",
    data[0x0], data[0x1], data[0x2], data[0x3],
    data[0x4], data[0x5], data[0x6], data[0x7],
    data[0x8], data[0x9], data[0xa], data[0xb],
    data[0xc], data[0xd], data[0xe], data[0xf],
    function, ttl, hops, size);
#endif
  
  if ((function != GNET_P_PING && function != GNET_P_PONG)) {
    DBUG("Not a ping or pong");
    return;
  }
  /* TTL of 0 or 1 should be OK */
  if (ttl > 1) {
    DBUG("Invalid TTL");
    return;
  }
  if (hops != 0) {
    DBUG("Invalid Hops");
    return;
  }
  if (size + 23 != len) {
    DBUG("Invalid size (size=%d, len=%d)", (int) size, (int) len);
    return;
  }
  
  switch ((gnet_p_t) function) {
  case GNET_P_PING:
    p = &data[23];
#if 0
    if (len == 23)
      DBUG("Got bare ping");
#endif
    break;

  case GNET_P_PONG:
    {
      unsigned long files, kbs;
      in_addr_t ip;
      
      if (len < 37)
        DBUG("Pong is too small");

      pong_port = peek_le16(&data[23]);
      memcpy(&ip, &data[25], sizeof ip);
      pong_addr = 0 != ip ? net_addr_set_ipv4(ip) : net_addr_unspecified;
      files = peek_le32(&data[29]);
      kbs = peek_le32(&data[33]);
      print_ipv4_addr(addr_buf, sizeof addr_buf, ip);
      DBUG("Got %spong (ip=%s, port=%u, files=%lu, kbs=%lu)",
        len == 37 ? "bare " : "", addr_buf, pong_port, files, kbs);
      
      p = &data[37];
    }
    break;
    
  case GNET_P_PUSH:
  case GNET_P_QASK:
  case GNET_P_QHIT:
    break;
  } 
  len -= (p - data);

  if (!ggep_decode(&gtx, p, len)) {
    DBUG("No GGEP data in packet");
    return;
  }
  
  for (;;) {
    char ggep_buf[4096];
    int ret;

    ret = ggep_next(&gtx, id_name);
    if (0 == ret)
      break;
    
    if (-1 == ret) {
      DBUG("Could not get next GGEP block");
      break;
    }
    
    id = ggep_map_id_name(id_name, NULL);
    if (GGEP_ID_INVALID == id) {
      DBUG("Unknown GGEP ID: \"%s\"", id_name);
      break;
    }

    data_len = ggep_data(&gtx, cast_to_void(&p), ggep_buf, sizeof ggep_buf);
    if ((ssize_t) -1 == (ssize_t) data_len) {
      WARN("Decompression of GGEP \"%s\" failed", id_name);
      continue;
    }
    
    switch (id) {
    case GGEP_ID_UDPHC:
      {

        host_is_udphc = true;
        if (data_len == 0) {
          DBUG("UDPHC: No hostname given");
        } else {
          char host[256 + 1], *ep;
          size_t size = sizeof host;
          bool truncated = false;

          if (size >= data_len + 1) {
            size = data_len + 1;
          } else {
            DBUG("UDPHC: Hostname is too long");
            truncated = true;
          }

          ep = append_string(host, &size, p);
          if (size > 1) {
            DBUG("UDPHC: Hostname contained NUL (size=%d)", (int) size);
          }
          DBUG("UDPHC: Hostname=\"%s%s\"", host, truncated ? " ..." : "");
        }
      }
      break;

    case GGEP_ID_IPP:
      if (0 == data_len || 0 != (data_len % 6)) {
        DBUG("IPP payload length (%d) is not a multiple of 6",
          (int) data_len);
      } else {
#if 1 
        DBUG("%d peers in IPP", (int) (data_len / 6));
#endif
        while (data_len) {
          in_addr_t ip;
          in_port_t port;
            
          memcpy(&ip, p, sizeof ip);
          p += sizeof ip;
          port = peek_le16(p);
          p += 2;
          
          print_ipv4_addr(addr_buf, sizeof addr_buf, ip);
#if 1
          DBUG("IPP: %s:%u", addr_buf, port);
#endif
          
          assert(data_len >= 6);
          data_len -= 6;

          if (ipp_peers.n < ARRAY_LEN(ipp_peers.addr)) {
            size_t i = ipp_peers.n++;
            
            ipp_peers.addr[i].addr = net_addr_set_ipv4(ip);
            ipp_peers.addr[i].port = port;
          }
        }
      }
      break;

    case GGEP_ID_SCP:
      if (has_scp)
        WARN("Multiple GGEP SCP blocks");

      has_scp = true;
      if (data_len > 0) {
        DBUG("SCP: Prefers free %s slots", *p & 1 ? "ultrapeer" : "leaf");
      } else {
        DBUG("SCP: No slot preferences");
      }
      break;

    case GGEP_ID_PHC:
      if (0 == data_len) {
        DBUG("PHC: No payload");
      }
      
      while (data_len > 0) {
        char buf[512];
        uint8_t *q;
        size_t left, len;

        assert(data_len <= INT_MAX);
        q = memchr(p, '\n', data_len);
        if (q) {
          *q++ = '\0';
        } else {
          q = &p[data_len];
        }
        len = q - p;
        assert(data_len >= len);
        data_len -= len;
        left = MIN(sizeof buf, (len + 1));
        (void) append_string(buf, &left, p);
        p = q;

        DBUG("PHC: \"%s\"", buf);
      }
      break;

    case GGEP_ID_DU:
      if (data_len > 3 || data_len < 1) {
        DBUG("Invalid length of DU payload: data_len=%d", (int) data_len);
      } else {
        size_t i = data_len;
        unsigned long uptime = 0;

        do {
          uptime = (uptime << 8) | p[--i];
        } while (i != 0);
#if 1
        DBUG("Daily Uptime: %lu seconds", uptime);
#endif
      }
      break;

    case GGEP_ID_LOC:
      if (data_len >= 2) {
        char loc[32], *q;
        size_t left = sizeof loc - 1;
       
        q = append_escaped_chars(loc, &left, p, data_len);
        *q = '\0';

#if 1 
        DBUG("Locale preference: \"%s\"", loc);
#endif
      } else {
        DBUG("Invalid length for LOC (len=%d)", (int) data_len);
      }
      break;

    case GGEP_ID_GUE:
#if 1
      DBUG("Node supports GUESS");
#endif
      if (data_len > 0) {
        char buf[64], *q;
        size_t left = sizeof buf - 1;

        q = append_escaped_chars(buf, &left, p, data_len);
        *q = '\0';
        
#if 1
        DBUG("GUE payload: \"%s\"", buf);
#endif
      }
      break;

    case GGEP_ID_VC:
      if (has_vc)
        WARN("Multiple GGEP VC blocks");
      
      if (data_len < 5) {
        DBUG("Invalid length for VC (len=%d)", (int) data_len);
      }
      
      if (data_len >= 5 && !has_vc) {
        char *q;
        size_t left = sizeof vendor - 1;

        memcpy((char *) &vc, p, 4);
        q = append_escaped_chars(vendor, &left, p, data_len);
        *q = '\0';
        DBUG("Vendor: \"%s\" char=%02x", vendor, p[4]);
      }

      has_vc = true;
      break;

    case GGEP_ID_UP:
      if (data_len == 3) {
#if 1
        DBUG("Free slots: %u/%u (UP/Leaf) char=%02x", p[1], p[2], p[0]);
#endif
      } else {
        DBUG("Invalid length for UP (len=%d)", (int) data_len);
      }
      break;

    default:
      DBUG("Unknown GGEP ID: \"%s\"", id_name);
    }

  }

  switch (function) {
  case GNET_P_PONG: 
    {
      guid_t guid;
      bool is_magic, ip_match = true, port_match = true, is_dupe = false;
      ping_info_t *pi = NULL;
      void *udata;
      struct timeval tv;
      int rtt = 0;

      memcpy(guid.u8, data, sizeof guid.u8);
      is_magic = guid_is_magic(&guid, &udata);
      if (!is_magic) {
        DBUG("Non-magic GUID");
      } else {
        assert(udata != NULL);

        pi = udata;
        gettimeofday(&tv, NULL);
        rtt = DIFFTIMEVAL(&tv, &pi->tv) / 1000;
        ip_match = net_addr_equal(pi->addr, sender_addr) ||
          net_addr_equal(pi->addr, net_addr_unspecified);
        port_match = pi->port == sender_port;
        is_dupe = pi->replied;
        pi->replied = false;

        print_net_addr_port(addr_buf, sizeof addr_buf, pi->addr, pi->port);
#if 1 
        DBUG("Extracted: RTT=%dms Peer=%s \t%s%s",
          rtt, addr_buf,
          ip_match && port_match ? "" : " MISMATCH!",
          is_dupe ? " DUPE!" : "");
#endif
      }

      /* Note, we don't ``pi'' here because guppy is not a long-lived
       * process and we might want to catch resent pongs. */
      
      if (
        !ip_match ||
        (
          !net_addr_equal(pong_addr, sender_addr) &&
          !net_addr_equal(pong_addr, net_addr_unspecified)
        )
      ) {
        DBUG("Address mismatch");
      } else if (!port_match || sender_port != pong_port) {
        DBUG("Port mismatch");
      } else if (is_dupe) {
        DBUG("Sender resent pong(?)");
      } else if (!is_magic) {
        DBUG("GUID is not magic");
      } else if (rtt > 3 * 60000) {
        DBUG("Pong is out of time");
      } else {
        if (pi)
          pi->replied = true;

        udp_pong_callback(sender_addr, sender_port, &ipp_peers);
      }
    }
    break;
    
  case GNET_P_PING:
    {
      guid_t guid;

      memcpy(&guid.u8, data, sizeof guid);

      if (!has_scp) {
        /* Nothing to do */
      } else if (sender_port < 1024) {
        DBUG("Not ponging privileged port");
      } else if (net_addr_equal(sender_addr, udp_con->get_addr(udp_con))) {
        DBUG("Ping from myself(?)");
      } else if (guid_is_bogus(&guid)) {
        DBUG("Not ponging bogus GUID");
      } else if (net_addr_is_private(sender_addr)) {
        DBUG("Not ponging private address");
      }
    }
    break;

  case GNET_P_PUSH:
  case GNET_P_QASK:
  case GNET_P_QHIT:
    break;
  }
  
}

static void
handle_udp(connection_t *c, connection_event_t ev)
{
  int fd = c->get_fd(c);
  ssize_t ret;
  char addr_buf[NET_ADDR_PORT_BUFLEN];
  char buf[4096];

  if (ev & CONN_EV_READ) {
    static const struct sockaddr_storage zero_from;
    struct sockaddr_storage from;
    socklen_t len = sizeof from;
    net_addr_t addr;
    in_port_t port;
    
    from = zero_from;
    ret = recvfrom(fd, buf, sizeof buf - 1, 0, cast_to_void(&from), &len);

    switch (from.ss_family) {
    case AF_INET:
      {
        const struct sockaddr_in *sin = cast_to_void(&from);

        addr = net_addr_set_ipv4(sin->sin_addr.s_addr);
        port = ntohs(sin->sin_port);
        print_net_addr_port(addr_buf, sizeof addr_buf, addr, port);
      }
      break;
      
#ifdef HAVE_IPV6_SUPPORT
    case AF_INET6:
      {
        const struct sockaddr_in6 *sin6 = cast_to_void(&from);

        addr = net_addr_set_ipv6(sin6->sin6_addr.s6_addr);
        port = ntohs(sin6->sin6_port);
        print_net_addr_port(addr_buf, sizeof addr_buf, addr, port);
      }
      break;
#endif /* HAVE_IPV6_SUPPORT */

    default:
      {
        size_t size = sizeof addr_buf;
        
        addr = net_addr_unspecified;
        port = 0;
        append_string(addr_buf, &size, "<none>");
      }
    }


    if (ret == 0) {
      DBUG("recvfrom() returned zero (from=%s)", addr_buf);
    } else if (ret == (ssize_t) -1) {
      DBUG("recvfrom() failed (from=%s): %s", addr_buf, compat_strerror(errno));
    } else {
      assert(ret >= 0 && (size_t) ret < sizeof buf);
      buf[ret] = '\0';
      DBUG("recvfrom()=%d (from=%s)", (int) ret, addr_buf);

      handle_packet(c, buf, ret, addr, port);
    }
  }
}

int
udp_init(server_t *server, connection_t *udp_c)
{
  int opt;
  socklen_t optlen = sizeof opt;
  
  assert(server != NULL);
  assert(udp_c != NULL);
  udp_con = udp_c;
  
  if (guid_init(1000000)) {
    CRIT("guid_init() failed");
    return -1;
  }

  if (NULL == (mc_ping_info = mem_new(sizeof(struct ping_info), "ping_info"))) {
    CRIT("mem_new() failed");
    return -1;
  }

  udp_con->set_blocking(udp_con, false);
  udp_con->set_rcvlowat(udp_con, 23); /* Minimum Gnutella packet size */

  opt = 2 * 64 * 1000;
  if (
    setsockopt(udp_con->get_fd(udp_con), SOL_SOCKET, SO_RCVBUF, &opt, optlen)
  ) {
    WARN("setsockopt() for SO_RCVBUF failed: %s", compat_strerror(errno));
  }
  opt = 2 * 64 * 1000;
  if (
    setsockopt(udp_con->get_fd(udp_con),
        SOL_SOCKET, SO_SNDBUF, &opt, optlen)
  ) {
    WARN("setsockopt() for SO_SNDBUF failed: %s", compat_strerror(errno));
  }
     

  udp_con->set_event_cb(udp_con, &handle_udp);
  server->add_client(server, udp_con);
  server->watch_client(server, udp_con, CONN_EV_READ);

  return 0;
}

/* vi: set ai et ts=2 sts=2 sw=2 cindent: */
